----------------------------------------------------------------------------------
-- Company: Univ Lyon, UJM-Saint-Etienne, CNRS, Laboratoire Hubert Curien UMR 5516, F-42023, SAINT-ETIENNE, France
-- Creator: Cédric Marchand 
-- 
-- Create Date:    18:07:09 11/11/2015 
-- Design Name: 
-- Module Name:    Datapath - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

LIBRARY work;
USE work.API_pkg.ALL;

entity Datapath is
  PORT(
    clk_i : in std_logic;
    --Input
    data_i    : in std_logic_vector(7 downto 0);
    key_i     : in std_logic_vector(7 downto 0);
    --Control
    mode_i    : in std_logic;
    key_sel   : in std_logic_vector(1 downto 0);
    data_sel  : in std_logic_vector(1 downto 0);
    count_r   : in std_logic_vector(4 downto 0);
    --Output
    data_o    : out std_logic_vector(7 downto 0)
  );
end Datapath;

architecture Behavioral of Datapath is

  COMPONENT klein_keyschedule
	PORT(
		key_in : IN std_logic_vector(79 downto 0);
		i : IN std_logic_vector(4 downto 0);          
		key_out : OUT std_logic_vector(79 downto 0)
		);
	END COMPONENT;
  
  COMPONENT klein_keyschedule_d
	PORT(
		key_in : IN std_logic_vector(79 downto 0);
		i : IN std_logic_vector(4 downto 0);          
		key_out : OUT std_logic_vector(79 downto 0)
		);
	END COMPONENT;
  
  --cipher
  COMPONENT klein_subnibbles
	PORT(
		state_i : IN std_logic_vector(63 downto 0);          
		state_o : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT klein_rotnibbles
	PORT(
		mode_i : IN std_logic;
		b_in : IN std_logic_vector(63 downto 0);          
		c_out : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT klein_mixnibbles
	PORT(
		state_in : IN std_logic_vector(63 downto 0);          
		state_out : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  COMPONENT klein_mixnibbles_d
	PORT(
		state_in : IN std_logic_vector(63 downto 0);          
		state_out : OUT std_logic_vector(63 downto 0)
		);
	END COMPONENT;
  
  --signals
  --regsiters
  signal key_reg      : std_logic_vector(key_size-1 downto 0);
  signal state_reg    : std_logic_vector(state_size-1 downto 0);
  --process signals
  signal key_o        : std_logic_vector(key_size-1 downto 0);
  signal keyd_o       : std_logic_vector(key_size-1 downto 0);
  signal round_o      : std_logic_vector(state_size-1 downto 0);
  signal addroundk_o  : std_logic_vector(state_size-1 downto 0);
  signal addroundkd_o : std_logic_vector(state_size-1 downto 0);
  signal sub_i        : std_logic_vector(state_size-1 downto 0);
  signal sub_o        : std_logic_vector(state_size-1 downto 0);
  signal rot_i        : std_logic_vector(state_size-1 downto 0);
  signal rot_o        : std_logic_vector(state_size-1 downto 0);
  signal mix_o        : std_logic_vector(state_size-1 downto 0);
  signal mixd_o       : std_logic_vector(state_size-1 downto 0);
  
begin
  --key process
  K_proc : process(clk_i, key_sel)
  begin
    if rising_edge(clk_i) then
      if key_sel = "01" then
        key_reg <= key_i & key_reg(key_size-1 downto 8);
      elsif key_sel = "10" then
        key_reg <= key_o;
      elsif key_sel = "11" then
        key_reg <= keyd_o;
      else
        key_reg <= key_reg;
      end if;
    end if;
  end process;
  
  --data process
  D_proc : process(clk_i, data_sel)
  begin
    if rising_edge(clk_i) then
      if data_sel = "01" then
        state_reg <=  state_reg(state_size-9 downto 0) & data_i;
      elsif data_sel = "10" then
        state_reg <=  round_o;
      elsif data_sel = "11" then
        state_reg <= addroundk_o;
      else
        state_reg <= state_reg;
      end if;
    end if;
  end process;
  
  data_o <= state_reg(59 downto 56) & state_reg(63 downto 60); -- & state_reg(7 downto 4);

  -- cipher component
  --keyschedule
  Inst_klein_keyschedule: klein_keyschedule PORT MAP(
		key_in => key_reg,
		i => count_r,
		key_out => key_o
	);
  
  Inst_klein_keyschedule_d: klein_keyschedule_d PORT MAP(
		key_in => key_reg,
		i => count_r,
		key_out => keyd_o
	);
  
  --datapath
  --addroundkey
  addroundk_o <= state_reg xor key_reg(63 downto 0);
  
  with mode_i select
    sub_i <=
      addroundk_o when '0',
      rot_o when others;
      
  Inst_klein_subnibbles: klein_subnibbles PORT MAP(
		state_i => sub_i,
		state_o => sub_o
	);
  
  with mode_i select
    rot_i <=
      sub_o   when '0',
      mixd_o  when others;
      
  Inst_klein_rotnibbles: klein_rotnibbles PORT MAP(
		mode_i => mode_i,
		b_in => rot_i,
		c_out => rot_o
	);
  
  Inst_klein_mixnibbles: klein_mixnibbles PORT MAP(
		state_in => rot_o,
		state_out => mix_o
	);
  
  Inst_klein_mixnibbles_d: klein_mixnibbles_d PORT MAP(
		state_in => addroundk_o,
		state_out => mixd_o
	);
  
  with mode_i select
    round_o <=
      mix_o when '0',
      sub_o when others;
  
end Behavioral;

