--Klein, keyschedule (for decryption)

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

ENTITY klein_keyschedule_d IS
  PORT (
    key_in  : IN STD_LOGIC_VECTOR(79 DOWNTO 0);
    i       : IN STD_LOGIC_VECTOR(4 DOWNTO 0);
    key_out : OUT STD_LOGIC_VECTOR(79 DOWNTO 0)
    );
END klein_keyschedule_d;

ARCHITECTURE rtl OF klein_keyschedule_d IS

  SIGNAL a : STD_LOGIC_VECTOR(39 DOWNTO 0);
  SIGNAL b : STD_LOGIC_VECTOR(39 DOWNTO 0);
  
  SIGNAL tmp : STD_LOGIC_VECTOR(4 DOWNTO 0);
  
  COMPONENT klein_sbox IS
    PORT (
      a_in  : IN STD_LOGIC_VECTOR(3 DOWNTO 0);
      a_out : OUT STD_LOGIC_VECTOR(3 DOWNTO 0)
      );
  END COMPONENT;
  
BEGIN

  tmp <= (key_in(16) & key_in(23 DOWNTO 20)) XOR STD_LOGIC_VECTOR(UNSIGNED(i));

  a(7 DOWNTO 0)   <= key_in(39 DOWNTO 32);
  a(27 DOWNTO 8)  <= key_in(19 DOWNTO 17) & tmp(4) & key_in(15 DOWNTO 0);
  a(31 DOWNTO 28) <= tmp(3 downto 0);
  a(39 downto 32) <= key_in(31 DOWNTO 24);

  b(7 DOWNTO 0)   <= key_in(79 DOWNTO 72);
  b(15 DOWNTO 8)  <= key_in(47 DOWNTO 40);
  b(39 DOWNTO 32) <= key_in(71 DOWNTO 64);

  i_klein_sbox1 : klein_sbox
    PORT MAP (
      a_in  => key_in(51 DOWNTO 48),
      a_out => b(19 DOWNTO 16)
      );

  i_klein_sbox2 : klein_sbox
    PORT MAP (
      a_in  => key_in(55 DOWNTO 52),
      a_out => b(23 DOWNTO 20)
      );

  i_klein_sbox3 : klein_sbox
    PORT MAP (
      a_in  => key_in(59 DOWNTO 56),
      a_out => b(27 DOWNTO 24)
      );

  i_klein_sbox4 : klein_sbox
    PORT MAP (
      a_in  => key_in(63 DOWNTO 60),
      a_out => b(31 DOWNTO 28)
      );

  key_out(39 DOWNTO 0)  <= a XOR b;
  key_out(79 DOWNTO 40) <= a;

END rtl;


