--Klein, MixNibbles step (encryption)

LIBRARY ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

ENTITY klein_mixnibbles IS
  PORT (
    state_in  : IN STD_LOGIC_VECTOR(63 DOWNTO 0);
    state_out : OUT STD_LOGIC_VECTOR(63 DOWNTO 0)
    );
END klein_mixnibbles;

ARCHITECTURE rtl OF klein_mixnibbles IS

  SIGNAL tmp  : STD_LOGIC_VECTOR(63 DOWNTO 0);
  SIGNAL oct0 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct1 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct2 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct3 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct4 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct5 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct6 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL oct7 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int1 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int2 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int3 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int4 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int5 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int6 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int7 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int8 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int9 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int10 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int11 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int12 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int13 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int14 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int15 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL int16 : STD_LOGIC_VECTOR(7 DOWNTO 0);
  
  COMPONENT klein_xtimes IS
    PORT (
      x_in  : IN STD_LOGIC_VECTOR(7 DOWNTO 0);
      x_out : OUT STD_LOGIC_VECTOR(7 DOWNTO 0)
      );
  END COMPONENT;
  
BEGIN

  oct0 <= state_in(3 DOWNTO 0) & state_in(7 DOWNTO 4);
  oct1 <= state_in(11 DOWNTO 8) & state_in(15 DOWNTO 12);
  oct2 <= state_in(19 DOWNTO 16) & state_in(23 DOWNTO 20);
  oct3 <= state_in(27 DOWNTO 24) & state_in(31 DOWNTO 28);
  oct4 <= state_in(35 DOWNTO 32) & state_in(39 DOWNTO 36);
  oct5 <= state_in(43 DOWNTO 40) & state_in(47 DOWNTO 44);
  oct6 <= state_in(51 DOWNTO 48) & state_in(55 DOWNTO 52);
  oct7 <= state_in(59 DOWNTO 56) & state_in(63 DOWNTO 60);

  int1 <= oct0 XOR oct1;
    
  i_klein_xtimes1 : klein_xtimes
    PORT MAP (
      x_in  => int1,
      x_out => int2
      );

  tmp(7 DOWNTO 0) <= int2 XOR oct1 XOR oct2 XOR oct3;

  int3 <= oct1 XOR oct2;

  i_klein_xtimes2 : klein_xtimes
    PORT MAP (
      x_in  => int3,
      x_out => int4
      );
  
  tmp(15 DOWNTO 8) <= int4 XOR oct0 XOR oct2 XOR oct3;

  int5 <= oct2 XOR oct3;

  i_klein_xtimes3 : klein_xtimes
    PORT MAP (
      x_in  => int5,
      x_out => int6
      );

  tmp(23 DOWNTO 16) <= int6 XOR oct0 XOR oct1 XOR oct3;

  int7 <= oct3 XOR oct0;

  i_klein_xtimes4 : klein_xtimes
    PORT MAP (
      x_in  => int7,
      x_out => int8
      );

  tmp(31 DOWNTO 24) <= int8 XOR oct0 XOR oct1 XOR oct2;

  int9 <= oct4 XOR oct5;

  i_klein_xtimes5 : klein_xtimes
    PORT MAP (
      x_in  => int9,
      x_out => int10
      );

  tmp(39 DOWNTO 32) <= int10 XOR oct5 XOR oct6 XOR oct7;

  int11 <= oct5 XOR oct6;

  i_klein_xtimes6 : klein_xtimes
    PORT MAP (
      x_in  => int11,
      x_out => int12
      );

  tmp(47 DOWNTO 40) <= int12 XOR oct4 XOR oct6 XOR oct7;

  int13 <= oct6 XOR oct7;

  i_klein_xtimes7 : klein_xtimes
    PORT MAP (
      x_in  => int13,
      x_out => int14
      );

  tmp(55 DOWNTO 48) <= int14 XOR oct4 XOR oct5 XOR oct7;

  int15 <= oct7 XOR oct4;

  i_klein_xtimes8 : klein_xtimes
    PORT MAP (
      x_in  => int15,
      x_out => int16
      );

  tmp(63 DOWNTO 56) <= int16 XOR oct4 XOR oct5 XOR oct6;

  state_out(3 DOWNTO 0)   <= tmp(7 DOWNTO 4);
  state_out(7 DOWNTO 4)   <= tmp(3 DOWNTO 0);
  state_out(11 DOWNTO 8)  <= tmp(15 DOWNTO 12);
  state_out(15 DOWNTO 12) <= tmp(11 DOWNTO 8);
  state_out(19 DOWNTO 16) <= tmp(23 DOWNTO 20);
  state_out(23 DOWNTO 20) <= tmp(19 DOWNTO 16);
  state_out(27 DOWNTO 24) <= tmp(31 DOWNTO 28);
  state_out(31 DOWNTO 28) <= tmp(27 DOWNTO 24);
  state_out(35 DOWNTO 32) <= tmp(39 DOWNTO 36);
  state_out(39 DOWNTO 36) <= tmp(35 DOWNTO 32);
  state_out(43 DOWNTO 40) <= tmp(47 DOWNTO 44);
  state_out(47 DOWNTO 44) <= tmp(43 DOWNTO 40);
  state_out(51 DOWNTO 48) <= tmp(55 DOWNTO 52);
  state_out(55 DOWNTO 52) <= tmp(51 DOWNTO 48);
  state_out(59 DOWNTO 56) <= tmp(63 DOWNTO 60);
  state_out(63 DOWNTO 60) <= tmp(59 DOWNTO 56);

  
END rtl;

